/*
 * Copyright 2001-2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.commons.net.nntp; 

/**
 * A placeholder interface for threadable message objects
 * Author: Rory Winston <rwinston@checkfree.com>
 *
 */
public interface Threadable {
	public boolean isDummy();
	public String messageThreadId();
	public String[] messageThreadReferences();
	public String simplifiedSubject();
	public boolean subjectIsReply();
	public void setChild(Threadable child); 
	public void setNext(Threadable next);
	public Threadable makeDummy();
}